@@ -77,6 +77,15 @@ pub enum CachePolicy {
     /// the FUSE client (i.e., the file system does not have exclusive access to the directory).
     Never,
 
+    /// This is almost same as Never, but it allows page cache of directories, dentries and attr
+    /// cache in guest. In other words, it acts like cache=never for normal files, and like
+    /// cache=always for directories, besides, metadata like dentries and attrs are kept as well.
+    /// This policy can be used if:
+    /// 1. the client wants to use Never policy but it's performance in I/O is not good enough
+    /// 2. the file system has exclusive access to the directory
+    /// 3. cache directory content and other fs metadata can make a difference on performance.
+    Metadata,
+
     /// The client is free to choose when and how to cache file data. This is the default policy and
     /// uses close-to-open consistency as described in the enum documentation.
     #[default]
@@ -93,10 +102,11 @@ impl FromStr for CachePolicy {
     type Err = &'static str;
 
     fn from_str(s: &str) -> Result<Self, Self::Err> {
-        match s {
-            "never" | "Never" | "NEVER" => Ok(CachePolicy::Never),
-            "auto" | "Auto" | "AUTO" => Ok(CachePolicy::Auto),
-            "always" | "Always" | "ALWAYS" => Ok(CachePolicy::Always),
+        match &s.to_lowercase()[..] {
+            "never" => Ok(CachePolicy::Never),
+            "metadata" => Ok(CachePolicy::Metadata),
+            "auto" => Ok(CachePolicy::Auto),
+            "always" => Ok(CachePolicy::Always),
             _ => Err("invalid cache policy"),
         }
     }
@@ -792,6 +802,13 @@ impl PassthroughFs {
                 OpenOptions::DIRECT_IO,
                 flags & (libc::O_DIRECTORY as u32) == 0,
             ),
+            CachePolicy::Metadata => {
+                if flags & (libc::O_DIRECTORY as u32) == 0 {
+                    opts |= OpenOptions::DIRECT_IO;
+                } else {
+                    opts |= OpenOptions::CACHE_DIR | OpenOptions::KEEP_CACHE;
+                }
+            }
             CachePolicy::Always => {
                 opts |= OpenOptions::KEEP_CACHE;
                 if flags & (libc::O_DIRECTORY as u32) != 0 {
@@ -1435,6 +1452,7 @@ impl FileSystem for PassthroughFs {
         let mut opts = OpenOptions::empty();
         match self.cfg.cache_policy {
             CachePolicy::Never => opts |= OpenOptions::DIRECT_IO,
+            CachePolicy::Metadata => opts |= OpenOptions::DIRECT_IO,
             CachePolicy::Always => opts |= OpenOptions::KEEP_CACHE,
             _ => {}
         };
